#include "hplog.h"
#ifdef _REENTRANT
#include <pthread.h>
#endif
#include <unistd.h>
#include <sys/types.h>
#include <stdio.h>
#include <errno.h>
#include <stdarg.h>

const char *Log_filename = NULL;
int Log_generations = -1;
int Log_maxsize = 1000000;


static int pid_logd = 0;
static int fd_logd = -1;
static FILE *fp_logd = (FILE*)NULL;
#ifdef _REENTRANT
static pthread_mutex_t mtx_log = PTHREAD_MUTEX_INITIALIZER;
#endif

int LogOpen() {
	/*NOT thread-safe. In multithraded programs the main thread should
          call LogOpen() once and for all*/
	
	int fd[2];
	int rc;

	rc = pipe(fd);
	if(rc!=0)
		return -1;
	
	pid_logd = fork();
	switch(pid_logd) {
		case -1:
			return -1;
		case 0: {
			/*child*/
			int argc;
			char *argv[5];
			char argv_str[5][20];
			
			close(fd[1]);
			dup2(fd[0],0);
			
			argc=0;
			strcpy(argv_str[argc],"hplogd");
			argv[argc]=argv_str[argc];
			argc++;
			if(Log_generations>=0) {
				sprintf(argv_str[argc],"-g%d",Log_generations);
				argv[argc] = argv_str[argc];
				argc++;
			}
			if(Log_maxsize>=0) {
				sprintf(argv_str[argc],"-s%d",Log_maxsize);
				argv[argc] = argv_str[argc];
				argc++;
			}
			if(Log_filename) {
				argv[argc] = (char*)Log_filename;
				argc++;
			}
			argv[argc] = 0;
			execvp("hplogd", argv);
			/*What???*/
			fprintf(stderr,"hplog.cc: panic: execvp(\"hplogd\",...) failed, errno=%d", errno);
			return -1;
		}
		default:
			close(fd[0]);
			fd_logd = fd[1];
			fp_logd = fdopen(fd_logd, "w");
	}
	return 0;
}

int LogSend(const void *buf, int buflen) {
	int rc;
	if(buflen<=0) return 0;
	
	if(pid_logd==0) {
		rc = LogOpen();
		if(rc) return rc;
	}
#ifdef _REENTRANT
	pthread_mutex_lock(&mtx_log);
#endif
	rc = write(fd_logd, buf, buflen);
#ifdef _REENTRANT
	pthread_mutex_unlock(&mtx_log);
#endif
	if(rc!=buflen)
		return -1;
	return 0;
}

int LogSendf(const char *fmt, ...) {
	va_list ap;
	int rc;
	if(pid_logd==0) {
		rc = LogOpen();
		if(rc) return rc;
	}
	va_start(ap,fmt);
#ifdef _REENTRANT
	pthread_mutex_lock(&mtx_log);
#endif
	vfprintf(fp_logd,fmt,ap);
	fflush(fp_logd);
#ifdef _REENTRANT
	pthread_mutex_unlock(&mtx_log);
#endif
	va_end(ap);
	return 0;
}

int LogClose() {
	if(pid_logd!=0) {
		close(fd_logd);
		/*the child should terminate by itself...*/
	}
	return 0;
}

#if 0
/*Simple test program:*/
int main(void) {
	LogOpen();
	LogSend("Hello world!\n",13);
	LogSendf("A random number: %d\n",42);
	LogClose();
}
#endif
